import datetime as dt
from collections.abc import Iterator
from decimal import Decimal
from typing import Any

from pandas import DataFrame
from pandas import Series
from sqlalchemy.engine import Engine

from dycw_utilities.numpy import datetime64ns
from dycw_utilities.numpy import has_dtype
from dycw_utilities.pandas import Int64
from dycw_utilities.pandas import boolean
from dycw_utilities.pandas import string
from dycw_utilities.pandas import timestamp_to_date
from dycw_utilities.pandas import timestamp_to_datetime
from dycw_utilities.sqlalchemy import get_column_names
from dycw_utilities.sqlalchemy import get_columns
from dycw_utilities.sqlalchemy import get_table


def insert_dataframe(
    df: DataFrame, table_or_model: Any, engine: Engine, /
) -> None:
    """Insert a DataFrame into a database."""

    names = get_column_names(table_or_model)
    columns = get_columns(table_or_model)
    all_names_to_columns = dict(zip(names, columns))
    names_to_columns = {
        name: col
        for name, col in all_names_to_columns.items()
        if (df.columns == name).any()
    }
    nativized = (
        _nativize_column(df[name], column)
        for name, column in names_to_columns.items()
    )
    rows = zip(*nativized)
    dicts = [dict(zip(names_to_columns, row)) for row in rows]
    if len(dicts) >= 1:
        table = get_table(table_or_model)
        with engine.begin() as conn:
            _ = conn.execute(table.insert(), list(dicts))


def _nativize_column(series: Series, column: Any, /) -> Iterator[Any]:
    """Check the columns of the DataFrame form a subset of the columns of the
    table.
    """

    py_type = column.type.python_type
    if (
        (has_dtype(series, (bool, boolean)) and (py_type in {bool, int}))
        or (has_dtype(series, float) and (py_type is float))
        or (has_dtype(series, (int, Int64)) and (py_type is int))
        or (has_dtype(series, string) and (py_type is str))
    ):
        values = series.tolist()
    elif has_dtype(series, datetime64ns) and (py_type is dt.date):
        values = series.map(timestamp_to_date)
    elif has_dtype(series, datetime64ns) and (py_type is dt.datetime):
        values = series.map(timestamp_to_datetime)
    else:
        raise TypeError(f"Invalid types: {series}, {py_type}")
    for is_null, native in zip(series.isna(), values):
        yield None if is_null else native


def read_dataframe(sel: Any, engine: Engine, /) -> DataFrame:
    """Read a table from a database into a DataFrame."""

    with engine.begin() as conn:
        rows = conn.execute(sel).all()
    sel = {col.name: _get_dtype(col) for col in sel.selected_columns}
    return DataFrame(rows, columns=list(sel)).astype(sel)


def _get_dtype(column: Any, /) -> Any:
    """Get the mapping of names to dtypes."""

    py_type = column.type.python_type
    if py_type is bool:
        return boolean
    elif (py_type is float) or (py_type is Decimal):
        return float
    elif py_type is int:
        return Int64
    elif py_type is str:
        return string
    elif issubclass(py_type, dt.date):
        return datetime64ns
    else:
        raise TypeError(f"Invalid type: {py_type}")  # pragma: no cover



import datetime as dt
from typing import Any
from typing import Optional

from pandas import NaT
from pandas import Timestamp


Int64 = "Int64"
boolean = "boolean"
string = "string"


def timestamp_to_date(timestamp: Any, /) -> Optional[dt.date]:
    """Convert a timestamp to a date."""

    if (value := timestamp_to_datetime(timestamp)) is None:
        return None
    else:
        return value.date()


def timestamp_to_datetime(timestamp: Any, /) -> Optional[dt.datetime]:
    """Convert a timestamp to a datetime."""

    if isinstance(timestamp, Timestamp):
        return timestamp.to_pydatetime()
    elif timestamp is NaT:
        return None
    else:
        raise TypeError(f"Invalid type: {type(timestamp).__name__!r}")



import datetime as dt
from collections.abc import Callable
from os import getpid
from pathlib import Path
from typing import Optional
from typing import Union

from hypothesis.strategies import SearchStrategy
from sqlalchemy.engine import Engine

from dycw_utilities.hypothesis import draw_and_map
from dycw_utilities.hypothesis.tempfile import temp_dirs
from dycw_utilities.hypothesis.typing import MaybeSearchStrategy
from dycw_utilities.sqlalchemy import create_engine
from dycw_utilities.tempfile import TemporaryDirectory


def sqlite_engines(
    *,
    dir: MaybeSearchStrategy[Union[Path, TemporaryDirectory]] = temp_dirs(),
    post_init: Optional[Callable[[Engine], None]] = None,
) -> SearchStrategy[Engine]:
    """Strategy for generating SQLite engines."""

    return draw_and_map(_draw_sqlite_engines, dir, post_init=post_init)


def _draw_sqlite_engines(
    dir: Union[Path, TemporaryDirectory],
    /,
    *,
    post_init: Optional[Callable[[Engine], None]] = None,
) -> Engine:
    dir_use = dir if isinstance(dir, Path) else dir.name
    now = dt.datetime.now().strftime("%4Y%m%dT%H%M%S.%f")
    pid = getpid()
    path = dir_use.joinpath(f"db__{now}__{pid}.sqlite")
    engine = create_engine("sqlite", database=path.as_posix())
    if post_init is not None:
        post_init(engine)
    return engine



from re import search
from typing import Any
from typing import Optional

from sqlalchemy import Table
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.engine import URL
from sqlalchemy.engine import Engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.exc import OperationalError
from sqlalchemy.pool import NullPool
from sqlalchemy.pool import Pool


def create_engine(
    drivername: str,
    /,
    *,
    username: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    database: Optional[str] = None,
    poolclass: Optional[type[Pool]] = NullPool,
) -> Engine:
    """Create a SQLAlchemy engine."""

    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
    )
    return _create_engine(url, future=True, poolclass=poolclass)


def ensure_table_created(table_or_model: Any, engine: Engine, /) -> None:
    """Ensure a table is created."""

    table = get_table(table_or_model)
    try:
        with engine.begin() as conn:
            table.create(conn)
    # note that OperationalError < DatabaseError
    except OperationalError as error:
        # sqlite
        (msg,) = error.args
        if not search("table .* already exists", msg):  # pragma: no cover
            raise
    except DatabaseError as error:  # pragma: no cover
        # oracle
        (msg,) = error.args
        if not search(
            "ORA-00955: name is already used by an existing object", msg
        ):
            raise


def ensure_table_dropped(table_or_model: Any, engine: Engine, /) -> None:
    """Ensure a table is dropped."""

    table = get_table(table_or_model)
    try:
        with engine.begin() as conn:
            table.drop(conn)
    # note that OperationalError < DatabaseError
    except OperationalError as error:
        # sqlite
        (msg,) = error.args
        if not search("no such table", msg):  # pragma: no cover
            raise
    except DatabaseError as error:  # pragma: no cover
        # oracle
        (msg,) = error.args
        if not search("ORA-00942: table or view does not exist", msg):
            raise


def get_column_names(table_or_model: Any, /) -> list[str]:
    """Get the column names from a table or model."""

    return [col.name for col in get_columns(table_or_model)]


def get_columns(table_or_model: Any, /) -> list[Any]:
    """Get the columns from a table or model."""

    return list(get_table(table_or_model).columns)


def get_table(table_or_model: Any, /) -> Table:
    """Get the table from a ORM model."""

    if isinstance(table_or_model, Table):
        return table_or_model
    else:
        return table_or_model.__table__
